#define	ENTRY(name) \
	.globl name; \
	.align 16; \
	name:

	.macro PUSH_ERROR
	sub $4, %esp		/* space for error code */
	.endm
	
	.macro PUSH_TRAP_EBP trapno
	sub $4, %esp		/* space for trap number */
	push %ebp
	mov $\trapno, %ebp
	mov %ebp, -4(%esp)	/* save trap number on stack */
	.endm

	.macro PUSH_REGS
	push %edi
	push %esi
	push %edx
	push %ecx
	push %ebx
	push %eax
	.endm

	.macro POP_REGS
	pop %eax
	pop %ebx
	pop %ecx
	pop %edx
	pop %esi
	pop %edi
	pop %ebp
	.endm

	.macro RETURN
	add $8, %esp		/* remove error code & trap number */
	iret			/* jump back */
	.endm	

	.macro DO_TRAP trapno func
	PUSH_TRAP_EBP \trapno
	PUSH_REGS
	call \func
	POP_REGS
	RETURN
	.endm

/* ------------------------------------------------------------------ */

	.code32
	.text

/* handled by emu */

ENTRY(boot)
	lea stack_high, %esp	/* setup stack */
	sub $56, %esp		/* sizeof(struct regs_32) */
	call do_boot
	POP_REGS
	RETURN

ENTRY(illegal_instruction)
	PUSH_ERROR
	DO_TRAP 3 do_illegal_instruction

ENTRY(general_protection)
	DO_TRAP 13 do_general_protection

ENTRY(page_fault)
	DO_TRAP 14 do_page_fault

ENTRY(event_callback)
	PUSH_ERROR
	DO_TRAP -1 do_event_callback

ENTRY(xen_hypercall)
	PUSH_ERROR
	PUSH_TRAP_EBP -1
	PUSH_REGS
	call do_hypercall
	POP_REGS
	add $8, %esp		/* pop trapno & error code */
	cmp $-1, -4(%esp)
	jne 1f
	out %al, $0xe0          /* let userspace handle it */
1:
	iret

/* bounced to userspace */
	
ENTRY(division_by_zero)
	out %al, $0xe4
	iret

ENTRY(int_unknown)
	out %al, $0xef
	jmp loop

loop:
	hlt
	jmp loop

	.data
	.globl stack_low, stack_high
stack_low:
	.fill 4096,4,0
stack_high:
