#define	ENTRY(name) \
	.globl name; \
	.align 16; \
	name:

	.macro PUSH_ERROR
	sub $8, %rsp		/* space for error code */
	.endm
	
	.macro PUSH_TRAP_RBP trapno
	sub $8, %rsp		/* space for trap number */
	push %rbp
	mov $\trapno, %rbp
	mov %rbp, -8(%rsp)	/* save trap number on stack */
	.endm

	.macro PUSH_REGS
	push %rdi
	push %rsi
	push %r15
	push %r14
	push %r13
	push %r12
	push %r11
	push %r10
	push %r9
	push %r8
	push %rdx
	push %rcx
	push %rbx
	push %rax
	.endm

	.macro POP_REGS
	pop %rax
	pop %rbx
	pop %rcx
	pop %rdx
	pop %r8
	pop %r9
	pop %r10
	pop %r11
	pop %r12
	pop %r13
	pop %r14
	pop %r15
	pop %rsi
	pop %rdi
	pop %rbp
	.endm

	.macro RETURN
	add $16, %rsp		/* remove error code & trap number */
	iretq			/* jump back */
	.endm	

	.macro DO_TRAP trapno func
	PUSH_TRAP_RBP \trapno
	PUSH_REGS
	call \func
	POP_REGS
	RETURN
	.endm

/* ------------------------------------------------------------------ */

	.code64
	.text

/* handled by emu */

ENTRY(boot)
	lea stack_high(%rip), %rsp	/* setup stack */
	sub $176, %rsp			/* sizeof(struct regs_64) */
	call do_boot
	POP_REGS
	RETURN

ENTRY(illegal_instruction)
	PUSH_ERROR
	DO_TRAP 6 do_illegal_instruction

ENTRY(general_protection)
	DO_TRAP 13 do_general_protection

ENTRY(page_fault)
	DO_TRAP 14 do_page_fault

ENTRY(int_80)
	PUSH_ERROR
	DO_TRAP -1 do_int_80

ENTRY(event_callback)
	PUSH_ERROR
	DO_TRAP -1 do_event_callback

ENTRY(syscall)
	/* create interrupt/exception stack frame */
	movq %rsp, stack_high-16(%rip)
	leaq stack_high-16(%rip), %rsp
	push %r11                       /* rflags         */
	mov  $0xdeadbeef,%r11           /* emu C code must fixup cs & ss */
	movq %r11, stack_high-8(%rip)   /* ss             */
	push %r11                       /* cs             */
	push %rcx                       /* rip            */

	PUSH_ERROR
	PUSH_TRAP_RBP -1
	PUSH_REGS
	call do_syscall
	POP_REGS

	add $16, %rsp			/* remove error code + trapno */
	cmp $-1, -8(%rsp)
	je syscall_vmexit
	cmp $-2, -8(%rsp)
	je syscall_iretq

syscall_default:
	/* default sysret path */
	popq %rcx                       /* rip    */
	popq %r11                       /* cs     */
	popq %r11                       /* rflags */
	popq %rsp                       /* rsp    */
	sysretq

syscall_vmexit:
	/* bounce hypercall to userspace */
	popq %rcx                       /* rip    */
	popq %r11                       /* cs     */
	popq %r11                       /* rflags */
	popq %rsp                       /* rsp    */
	out %al, $0xe0                  /* let userspace handle it */
	sysretq

syscall_iretq:
	/* return from syscall via iretq */
	iretq

/* forwarded to userspace */

ENTRY(debug_trap)
	out %al, $0xec
	iretq

ENTRY(debug_int3)
	out %al, $0xec
	iretq

ENTRY(division_by_zero)
	out %al, $0xe4
	iretq

ENTRY(int_unknown)
	out %al, $0xef
	jmp loop

ENTRY(instructions)
	.byte 0
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	nop

loop:
	hlt
	jmp loop

	/* emu stack */
	.data
	.globl stack_low, stack_high
stack_low:
	.align 4096
	.fill 4096,4,0
stack_high:
